#include "usb_dev.h"
#include "usb_debug_hid.h"

#ifdef DEBUG_INTERFACE

uint8_t usb_debug_line_coding[7];
uint8_t usb_debug_line_rtsdtr;
volatile uint8_t usb_debug_transmit_flush_timer=0;

static usb_packet_t *debug_tx_packet=NULL;
static volatile uint8_t debug_tx_noautoflush=0;

#define TRANSMIT_FLUSH_TIMEOUT	5   /* in milliseconds */

// Maximum number of transmit packets to queue so we don't starve other endpoints for memory
#define debug_tx_packet_LIMIT 5

// When the PC isn't listening, how long do we wait before discarding data?  If this is
// too short, we risk losing data during the stalls that are common with ordinary desktop
// software.  If it's too long, we stall the user's program when no software is running.
#define TX_TIMEOUT_MSEC 50

#if F_CPU == 96000000
  #define TX_TIMEOUT (TX_TIMEOUT_MSEC * 596)
#elif F_CPU == 48000000
  #define TX_TIMEOUT (TX_TIMEOUT_MSEC * 428)
#elif F_CPU == 24000000
  #define TX_TIMEOUT (TX_TIMEOUT_MSEC * 262)
#endif

// When we've suffered the transmit timeout, don't wait again until the computer
// begins accepting data.  If no software is running to receive, we'll just discard
// data as rapidly as Serial.print() can generate it, until there's something to
// actually receive it.
static uint8_t transmit_previous_timeout=0;

// transmit a character.  0 returned on success, -1 on error
int usb_debug_putchar(uint8_t c)
{
	return usb_debug_write(&c, 1);
}

int usb_debug_write(const void *buffer, uint32_t size)
{
	uint32_t len;
	uint32_t wait_count;
	const uint8_t *src = (const uint8_t *)buffer;
	uint8_t *dest;

	debug_tx_noautoflush = 1;
	while (size > 0) {
		if (!debug_tx_packet) {
			wait_count = 0;
			while (1) {
				if (!usb_configuration) {
					debug_tx_noautoflush = 0;
					return -1;
				}
				if (usb_tx_packet_count(DEBUG_TX_ENDPOINT) < debug_tx_packet_LIMIT) {
					debug_tx_noautoflush = 1;
					debug_tx_packet = usb_malloc();
					if (debug_tx_packet) break;
					debug_tx_noautoflush = 0;
				}
				if (++wait_count > TX_TIMEOUT || transmit_previous_timeout) {
					transmit_previous_timeout = 1;
					return -1;
				}
			}
		}
		transmit_previous_timeout = 0;
		len = DEBUG_TX_SIZE - debug_tx_packet->index;
		if (len > size) len = size;
		dest = debug_tx_packet->buf + debug_tx_packet->index;
		debug_tx_packet->index += len;
		size -= len;
		while (len-- > 0) *dest++ = *src++;
		if (debug_tx_packet->index < DEBUG_TX_SIZE) {
			usb_debug_transmit_flush_timer = TRANSMIT_FLUSH_TIMEOUT;
		} else {
			debug_tx_packet->len = DEBUG_TX_SIZE;
			usb_debug_transmit_flush_timer = 0;
			usb_tx(DEBUG_TX_ENDPOINT, debug_tx_packet);
			debug_tx_packet = NULL;
		}
	}
	debug_tx_noautoflush = 0;
	return 0;
}

void usb_debug_flush_output(void)
{
    int i;
	if (!usb_configuration) return;
	//serial_print("usb_debug_flush_output\n");
	if (debug_tx_packet && debug_tx_packet->index > 0) {
		usb_debug_transmit_flush_timer = 0;
		for (i = debug_tx_packet->index; i < DEBUG_TX_SIZE; i++) {
			debug_tx_packet->buf[i] = 0;
		}
		debug_tx_packet->len = DEBUG_TX_SIZE;
		usb_tx(DEBUG_TX_ENDPOINT, debug_tx_packet);
		debug_tx_packet = NULL;
	}
	// while (usb_tx_byte_count(DEBUG_TX_ENDPOINT) > 0) ; // wait
}

void usb_debug_hid_flush_callback(void)
{
  int i;
	if (debug_tx_noautoflush) return;
	//serial_print("usb_flush_callback \n");
	for (i = debug_tx_packet->index; i < DEBUG_TX_SIZE; i++) {
		debug_tx_packet->buf[i] = 0;
	}
	usb_tx(DEBUG_TX_ENDPOINT, debug_tx_packet);
	debug_tx_packet = NULL;
	//serial_print("usb_flush_callback end\n");
}

void usb_debug_print(const char *p)
{
	while (*p) usb_debug_putchar(*p++);
}

static void usb_debug_phex1(uint8_t n)
{
	n &= 15;
	if (n < 10) {
		usb_debug_putchar('0' + n);
	} else {
		usb_debug_putchar('A' - 10 + n);
	}
}

void usb_debug_phex(uint8_t n)
{
	usb_debug_phex1(n >> 4);
	usb_debug_phex1(n);
}

void usb_debug_phex16(uint16_t n)
{
	usb_debug_phex(n >> 8);
	usb_debug_phex(n);
}

void usb_debug_phex32(uint32_t n)
{
	usb_debug_phex(n >> 24);
	usb_debug_phex(n >> 16);
	usb_debug_phex(n >> 8);
	usb_debug_phex(n);
}

#endif /* DEBUG_INTERFACE */
